# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the path splitting processor."""

import json
import os
import unittest
from unittest.mock import patch

from grove.models import ProcessorConfig
from grove.processors import split_path
from tests import mocks


class ProcessorPathMapperTestCase(unittest.TestCase):
    """Implements unit tests for the path splitting processor."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Setup the processor and associated configuration for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

        # Create a mapping compatible with the Okta system log fixture.
        self.processor = split_path.Handler(
            ProcessorConfig(
                name="Fan Out",
                processor="split_path",
                source="events",
            )
        )

    def test_split_path(self):
        """Ensure path splitting operates as expected."""
        # Load and process the fixture.
        entries = json.load(
            open(
                os.path.join(self.dir, "fixtures/gsuite/activities/001.json"),
                "r",
            )
        )

        # Confirm that the initial log entry has two entries.
        self.assertEqual(len(entries["items"]), 1)

        # Process a single log entry.
        records = self.processor.process(entries["items"][0])

        # Confirm that two records resulted.
        self.assertEqual(len(records), 2)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove connectors."""

import abc
import datetime
import hashlib
import json
import logging
import os
from typing import Any, Dict, List, Optional

import jmespath

from grove.__about__ import __version__
from grove.constants import (
    CACHE_KEY_LOCK,
    CACHE_KEY_POINTER,
    CACHE_KEY_POINTER_NEXT,
    CACHE_KEY_POINTER_PREV,
    CACHE_KEY_SEEN,
    CACHE_KEY_WINDOW_END,
    CACHE_KEY_WINDOW_START,
    CHRONOLOGICAL,
    DATESTAMP_FORMAT,
    DEFAULT_CACHE_HANDLER,
    DEFAULT_LOCK_DURATION,
    DEFAULT_OUTPUT_HANDLER,
    ENV_GROVE_CACHE_HANDLER,
    ENV_GROVE_LOCK_DURATION,
    ENV_GROVE_OUTPUT_HANDLER,
    LOCK_DATE_FORMAT,
    PLUGIN_GROUP_CACHE,
    PLUGIN_GROUP_OUTPUT,
    PLUGIN_GROUP_PROCESSOR,
    REVERSE_CHRONOLOGICAL,
)
from grove.exceptions import (
    AccessException,
    ConcurrencyException,
    ConfigurationException,
    DataFormatException,
    GroveException,
    NotFoundException,
    ProcessorError,
)
from grove.helpers import parsing, plugin
from grove.models import ConnectorConfig, OutputStream


class BaseConnector:
    NAME = "base"
    POINTER_PATH = "NOT_SET"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def __init__(self, config: ConnectorConfig, context: Dict[str, str]):
        """Sets up a Grove connector.

        :param config: A valid ConnectorConfig object containing information to use
            when configuring this connector.
        :param context: Contextual information relating to the current runtime.
        """
        self.logger = logging.getLogger(__name__)
        self.configuration = config
        self.runtime_context = context

        # 'Operations' are useful for APIs which have MANY different types of data which
        # can be returned but where only a small subset are pertinent. In this case, the
        # operation can be set in the configuration to instruct grove to only collect
        # those types of log data.
        self.key = self.configuration.key
        self.identity = self.configuration.identity
        self.operation = self.configuration.operation

        # Define contextual log data to be appended to all log messages.
        self.log_context = {
            "operation": self.operation,
            "connector": self.__class__.__module__,
            "identity": self.identity,
        }

        # Let the caller handle exceptions from failure to load handlers directly.
        self._cache = plugin.load_handler(
            os.environ.get(ENV_GROVE_CACHE_HANDLER, DEFAULT_CACHE_HANDLER),
            PLUGIN_GROUP_CACHE,
        )
        self._output = plugin.load_handler(
            os.environ.get(ENV_GROVE_OUTPUT_HANDLER, DEFAULT_OUTPUT_HANDLER),
            PLUGIN_GROUP_OUTPUT,
        )
        self._output.setup()

        # Processors are only setup once for each connector instance.
        self._processors = {}

        for processor in self.configuration.processors:
            try:
                self._processors[processor.name] = plugin.load_handler(
                    processor.processor,
                    PLUGIN_GROUP_PROCESSOR,
                    processor,
                )
            except ConfigurationException as err:
                raise ProcessorError(
                    f"Failed to initialise processor '{processor.name}' "
                    f"({processor.processor}). {err}",
                )

        # The time that our current lock expires, if we have one.
        self._lock_expiry: Optional[datetime.datetime] = None

        try:
            self._lock_duration = int(
                os.environ.get(ENV_GROVE_LOCK_DURATION, DEFAULT_LOCK_DURATION)
            )
        except ValueError as err:
            self.logger.warning(
                f"Lock duration ('{ENV_GROVE_LOCK_DURATION}') must be a number.",
                extra={"exception": err, **self.log_context},
            )

        # Determines if the start of a 'window' has been passed during a collection.
        # This is used to track windows which span multiple pages of results and is only
        # applicable for logs collected in reverse chronological order.
        self._window_passed = False
        self._window_start = str()
        self._window_end = str()

        # Paginated / chunked data needs an incrementing identifier to keep things
        # orderly.
        self._part = 0

        # Track the number of output logs by the configured output destination stream.
        # This allows statistics to be generated on deduplication, splitting, etc.
        self._saved = {}
        for descriptor, _ in self.configuration.outputs.items():
            self._saved[descriptor] = 0

        # Tracks hashes of unique log entries, keyed by their pointer value.
        self._hashes: Dict[str, set[str]] = {}

        # Pointers track the last collected record in order for collection to continue
        # at the correct place between runs. A "next" pointer is only applicable for
        # logs collected in reverse chronological order.
        self._pointer = str()
        self._pointer_next = str()
        self._pointer_previous = str()

    def run(self):
        """Connector entrypoint, called by the scheduler.

        Wraps collect to handle exceptions and logging for collection. This method
        should NOT be implemented by connectors as it is only intended to provide a
        consistent calling and error handling mechanism when connectors are executed.
        """
        # Acquire a lock first.
        try:
            self.lock()
        except ConcurrencyException as err:
            self.logger.warning(
                "Connector may already be running in another location.",
                extra={"exception": err, **self.log_context},
            )

        # Perform collection.
        try:
            self.collect()
        except GroveException as err:
            self.logger.error(
                "Connector was unable to complete collection successfully.",
                extra={"exception": err, **self.log_context},
            )
            self.unlock()
            return

        # Call the correct post run method.
        if self.LOG_ORDER == CHRONOLOGICAL:
            self._run_chronological()

        if self.LOG_ORDER == REVERSE_CHRONOLOGICAL:
            self._run_reverse_chronological()

        # TODO: The use of a context manager for lock management would be best.
        self.unlock()

    def _run_chronological(self):
        """Performs chronological specific post collection operations."""
        # TODO: Move to processor.
        try:
            self.logger.debug(
                "Saving deduplication hashes to cache.",
                extra=self.log_context,
            )
            self.save_hashes()
        except AccessException as err:
            self.logger.error(
                "Failed to save hashes to cache.",
                extra={"exception": err, **self.log_context},
            )
            return

    def _run_reverse_chronological(self):
        """Performs reverse chronological specific post collection operations."""
        # Swap pointers.
        try:
            self.pointer = self.pointer_next
            self.logger.debug(
                "Pointer successfully saved to cache.",
                extra={"pointer": self.pointer_next, **self.log_context},
            )
        except AccessException as err:
            self.logger.error(
                "Connector failed to save pointer, cannot continue.",
                extra={"exception": err, **self.log_context},
            )
            return
        except NotFoundException:
            self.logger.debug(
                "Skipping pointer swap and clean-up as there is no next-pointer.",
                extra={**self.log_context},
            )
            return

        # If the collection complete without error delete the window locations and
        # the next pointer from cache, if set.
        try:
            self.logger.debug(
                "Deleting window and next pointer after successful collection.",
                extra=self.log_context,
            )
            self._cache.delete(
                self.cache_key(CACHE_KEY_WINDOW_START),
                self.operation,
            )
            self._cache.delete(
                self.cache_key(CACHE_KEY_WINDOW_END),
                self.operation,
            )
            self._cache.delete(
                self.cache_key(CACHE_KEY_POINTER_NEXT),
                self.operation,
            )
        except AccessException as err:
            self.logger.error(
                "Failed to clean up windows and next pointer from cache.",
                extra={"exception": err, **self.log_context},
            )
            return

        # TODO: Move to processor.
        try:
            self.logger.debug(
                "Saving deduplication hashes to cache",
                extra=self.log_context,
            )
            self.save_hashes()
        except AccessException as err:
            self.logger.error(
                "Failed to save hashes to cache.",
                extra={"exception": err, **self.log_context},
            )
            return

    @abc.abstractmethod
    def collect(self):
        """Provides a stub for a connector to initiate a collection."""
        pass

    def process_and_write(self, entries: List[Any]):
        """Write log entries them to the configured output handler.

        :param entries: List of log entries to process.
        """
        # Allow failures to bubble all the way up and fail the run. If processing fails
        # we want to defer collection, to allow retry later. We always pass a copy of
        # the entries to prevent accidental overwriting of the collected raw data by
        # a processor.
        processed = self.process(entries)

        for descriptor, stream in self.configuration.outputs.items():
            # Ensure the output uses the correct stream.
            to_save = entries
            if stream == OutputStream.processed:
                to_save = processed

            number_of_entries = len(to_save)
            if number_of_entries < 1:
                self.logger.info(
                    "No log entries to output for stream, skipping.",
                    extra={
                        "stream": stream,
                        "descriptor": descriptor,
                        **self.log_context,
                    },
                )
                continue

            try:
                self._output.submit(
                    data=self._output.serialize(
                        data=to_save,
                        metadata=self.metadata(),
                    ),
                    part=self._part,
                    operation=self.operation,
                    connector=self.NAME,
                    identity=self.identity,
                    descriptor=descriptor,
                )

                # Update counters.
                self._saved[descriptor] += number_of_entries

                self.logger.info(
                    "Log submitted successfully to output.",
                    extra={
                        "part": self._part,
                        "stream": stream,
                        "descriptor": descriptor,
                        "entries": number_of_entries,
                        **self.log_context,
                    },
                )
            except AccessException as err:
                self.logger.error(
                    "Failed to write logs to output, cannot continue.",
                    extra={
                        "part": self._part,
                        "exception": err,
                        "stream": stream,
                        "descriptor": descriptor,
                        **self.log_context,
                    },
                )
                raise

    def save(self, entries: List[Any]):
        """Saves log entries, and updates the pointer in the cache.

        :param entries: List of log entries to save.
        """
        # TODO: Move deduplication into a processor.
        entries = self.deduplicate_by_hash(entries)

        if len(entries) < 1:
            self.logger.warning(
                "No log entries to save, skipping.", extra=self.log_context
            )
            return

        # Always refresh our lock while saving. This allows us to grab a new lock for
        # every page of data to try and prevent our lock expiring before we've performed
        # a full collection.
        #
        # Unlock is not called here, as it's performed by the caller.
        self.lock()

        if self.LOG_ORDER == CHRONOLOGICAL:
            self._save_chronological(entries)

        if self.LOG_ORDER == REVERSE_CHRONOLOGICAL:
            self._save_reverse_chronological(entries)

        self.finalize()

    def _save_chronological(self, entries: List[Any]):
        """Saves log entries when retrieved logs are in chronological order.

        :param entries: List of log entries to save.
        """
        # Pointers are extracted prior to processing as processing may modify the
        # structure, or remove entries entirely.
        newest = jmespath.search(self.POINTER_PATH, entries[-1])
        if newest is None:
            raise GroveException(
                f"Pointer path ({self.POINTER_PATH}) was not found in returned logs."
            )

        # Exceptions are allowed to bubble up here to ensure connectors exit on error,
        # rather than silently dropping batches of log entries.
        self.process_and_write(entries)

        # Once uploaded, then update the pointer. NOTE: There is an opportunity for
        # issues to occur between the output and pointer update which would lead to
        # duplicate data.
        try:
            self.pointer = newest
            self.logger.info(
                "Pointer successfully saved to cache.",
                extra={"pointer": newest, **self.log_context},
            )
        except AccessException as err:
            self.logger.error(
                "Failed to save pointer to cache, cannot continue.",
                extra={"exception": err, **self.log_context},
            )
            raise

        # Get ready for the next batch of candidate log entries (if required).
        self._part += 1

    def _save_reverse_chronological(self, candidates: List[Any]):  # noqa: C901
        """Save log entries when logs are in reverse chronological order.

        Data returned in reverse chronological order is more complicated to handle,
        as the data may be paginated, and the execution environment "unreliable". This
        results in more operations are against the cache.

        TODO: This method needs some love. It's very complex, and could do with either
        simplifying or breaking into a number of methods.

        :param candidates: List of log entries to save.
        """
        entries = []
        oldest = jmespath.search(self.POINTER_PATH, candidates[-1])
        newest = jmespath.search(self.POINTER_PATH, candidates[0])

        if oldest is None or newest is None:
            raise GroveException(
                f"Pointer path ({self.POINTER_PATH}) was not found in returned logs."
            )

        # If a window start is in the cache then a previous collection is incomplete.
        # We'll skip entries until we find our window, and then only collect entries
        # which are ON AND AFTER the end of the window. Any skipped entries will be
        # fetched by subsequent collections.
        incomplete_collection = False

        try:
            if self.window_start:
                incomplete_collection = True
        except NotFoundException:
            pass

        if incomplete_collection:
            for entry in candidates:
                current_pointer = jmespath.search(self.POINTER_PATH, entry)

                if current_pointer is None:
                    raise GroveException(
                        f"Pointer path ({self.POINTER_PATH}) not found in log entry."
                    )

                # We need to track FROM the window end, inclusive, to ensure that we
                # don't miss any logs. This is required in cases where the timestamp
                # granularity from the upstream API is only seconds (as always, we
                # prefer duplicates to missing logs).
                #
                # TODO: There is a condition here were duplicates are not considered.
                if not self._window_passed and str(current_pointer) == self.window_end:
                    self._window_passed = True

                # This is expensive, as every log entry will result in a write to the
                # cache. This is why reverse chronological ordering should be avoided
                # where possible.
                if self._window_passed:
                    entries.append(entry)
                    self.window_end = current_pointer

        # Explicit test here rather than 'else' as the previous block is a little long.
        if not incomplete_collection:
            # Save all data if we don't have an existing incomplete collection.
            for entry in candidates:
                entries.append(entry)

            # Track the window of this collection, and the next pointer.
            self._window_passed = True

            self.window_start = newest
            self.window_end = oldest
            self.pointer_next = newest

        if len(entries) < 1:
            return

        # Exceptions are allowed to bubble up here to ensure connectors exit on error,
        # rather than silently dropping batches of log entries.
        self.process_and_write(entries)

        # Get ready for the next block of entries (if required).
        self._part += 1

        # Save the new window geometry to cache but only AFTER data is saved, and only
        # save the window start when it's updated.
        if not incomplete_collection:
            self.save_window_start()

        self.save_window_end()

    def metadata(self) -> Dict[str, Any]:
        """Returns contextual metadata associated with this collection.

        :return: A dictionary of metadata for storing with log entries.
        """
        return {
            "connector": self.__class__.__module__,
            "identity": self.identity,
            "operation": self.operation,
            "pointer": self.pointer,
            "previous_pointer": self.pointer_previous,
            "collection_time": datetime.datetime.utcnow().strftime(DATESTAMP_FORMAT),
            "runtime": self.runtime_context,
            "version": __version__,
        }

    def cache_key(self, prefix: str = CACHE_KEY_POINTER) -> str:
        """Generates a cache key which uniquely identifies this connector.

        This includes the name and identity to allow multiple instances of the same
        connector to be used concurrently.

        :param prefix: A prefix for the cache key.

        :return: The constructed cache key.
        """
        # MD5 may not be cryptographically secure, but it works for our purposes. It's:
        #
        #   1) Short.
        #   2) Has a low chance of (non-deliberate) collisions.
        #   3) Able to be 'stringified' as hex, the character set of which is compatible
        #      with backends like DynamoDB.
        #
        return ".".join(
            [
                prefix,
                self.NAME,
                hashlib.md5(bytes(self.identity, "utf-8")).hexdigest(),  # noqa: S324
            ]
        )

    def hash_entry(self, entry: Any) -> str:
        """Serialise and hash the provided log entry.

        This is intended to produce a unique identifier for this event which may be used
        for deduplication.

        :param entry: The log entry to hash.

        :return: A hash of a log entry.
        """
        content = bytes(json.dumps(entry, separators=(",", ":")), "utf-8")

        return hashlib.md5(content).hexdigest()

    def hash_entries(self, entries: List[Any]) -> Dict[str, set[str]]:
        """Hashes a list of log entries.

        :param entries: List of log entries to hash

        :return: A dictionary containing a set of log hashes, keyed by the pointer of
            each event.
        """
        hashes: Dict[str, set[str]] = {}

        for entry in entries:
            # If we can't find a pointer in the log entry, just skip it.
            entry_pointer = jmespath.search(self.POINTER_PATH, entry)
            if not entry_pointer:
                continue

            if entry_pointer not in hashes:
                hashes[entry_pointer] = set()

            hashes[entry_pointer].add(self.hash_entry(entry))

        return hashes

    def deduplicate_by_hash(self, candidates: List[Any]):
        """Deduplicate log entries by their hash.

        This is performed by generating a hash of the log entry, and comparing these
        hashes against recently seen events in the cache with the same pointer value.
        If a value is in the cache, then the log entry will be discarded.

        Please note: This only applies to events which have a pointer value that matches
        the most recently saved. This is to prevent needing to keep large amounts of log
        entry hashes.

        :param candidates: A list of log entries to deduplicate.

        :return: A deduplicated list of log entries.
        """
        entries = []
        old_hashes = self.hashes
        new_hashes: Dict[str, set[str]] = {}

        # Check whether these log entries have been seen already.
        for candidate in candidates:
            candidate_hash = self.hash_entry(candidate)
            candidate_pointer = str(jmespath.search(self.POINTER_PATH, candidate))

            # Track this log entry's hash.
            if candidate_pointer not in new_hashes:
                new_hashes[candidate_pointer] = set()

            new_hashes[candidate_pointer].add(candidate_hash)

            # Check if the hash for this log entry is in the cache, and if so, skip it.
            try:
                if candidate_pointer in old_hashes:
                    if candidate_hash in old_hashes[candidate_pointer]:
                        continue
            except KeyError:
                pass

            entries.append(candidate)

        # Update known in-memory hashes to include our new entries.
        self.hashes = {**old_hashes, **new_hashes}

        return entries

    def deduplicate_by_pointer(self, entries: List[Any]):
        """Deduplicate log entries by pointer values.

        Deduplicates records which occur before or after a pointer on the current
        page - depending on whether log entries are in chronological or reverse
        chronological order. This enables deduplication of log events within the same
        page of results, and is intended to solve for cases where a provider's filters
        are not as granular as the pointer value.

        For example, some provider's only allow filtering on a date (YYYY-MM-DD) while
        returning log entries with timestamps that have millisecond precision.

        :param entries: A list of log entries to deduplicate.

        :return: A deduplicated list of log entries.
        """
        if self.LOG_ORDER == CHRONOLOGICAL:
            return self._deduplicate_by_pointer_chronological(entries)

        if self.LOG_ORDER == REVERSE_CHRONOLOGICAL:
            return self._deduplicate_by_pointer_reverse_chronological(entries)

    def _deduplicate_by_pointer_chronological(self, entries: List[Any]):
        """Deduplicates chronological log entries by their pointer.

        :param entries: A list of log entries to deduplicate.

        :return: A deduplicated list of log entries.
        """
        results = []
        pointer_passed = False

        for candidate in entries:
            candidate_pointer = str(jmespath.search(self.POINTER_PATH, candidate))

            if candidate_pointer == self.pointer:
                pointer_passed = True

            # Only track chronological records on and after the pointer.
            if pointer_passed:
                results.append(candidate)

        # If we never encountered the pointer, don't filter the records at all. This may
        # cause some duplicates if the pointer is on a subsequent page, but we always
        # prefer duplicates in these cases.
        if not pointer_passed:
            results = entries

        return results

    def _deduplicate_by_pointer_reverse_chronological(self, entries: List[Any]):
        """Deduplicates reverse chronological log entries by their pointer.

        :param entries: A list of log entries to deduplicate.

        :return: A deduplicated list of log entries.
        """
        results = []
        pointer_found = False
        pointer_passed = False

        for candidate in entries:
            candidate_pointer = jmespath.search(self.POINTER_PATH, candidate)

            if candidate_pointer == self.pointer:
                pointer_found = True

            if pointer_found and candidate_pointer != self.pointer:
                pointer_passed = True
                break

            if not pointer_passed:
                results.append(candidate)

        # If we never encountered the pointer, don't filter the records at all. This may
        # cause some duplicates if the pointer is on a subsequent page, but we always
        # prefer duplicates in these cases.
        if not pointer_passed:
            results = entries

        return results

    def process(self, entries: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Process log entries prior to saving.

        :param entries: A list of log entries to process.

        :return: A processed list of log entries.
        """
        # Shortcut where there are no processors configured.
        if len(self._processors) < 1:
            return []

        # As processors can modify the number of entries we need to do a few operations
        # here to handle additions and removals between iterations.
        processed = parsing.quick_copy(entries)

        for name, processor in self._processors.items():
            index = 0
            processed_size = len(processed)

            while index < processed_size:
                try:
                    processed[index:index] = processor.process(processed.pop(index))
                except Exception as err:
                    raise ProcessorError(
                        f"Fatal error in processor '{name}' ({processor}). {err}"
                    )

                # Adjust the loop index based on whether we've added or removed items.
                current_size = len(processed)

                if current_size < processed_size:
                    processed_size = current_size
                else:
                    index += 1

        return processed

    def finalize(self):
        """Performs final steps after each save operation has complete."""

        # Finalize all processors.
        for name, processor in self._processors.items():
            # Once again this exception handler is exceptionally (!) broad, to ensure
            # that any unhandled exception, including from downstream libraries, are
            # caught and handled consistently (except for BaseException derived).
            try:
                processor.finalize()
            except Exception as err:
                # As this runs after saving data and pointers, all we can really do is
                # log this and continue.
                self.logger.error(
                    "Processor failed during finalization.",
                    extra={
                        "identity": name,
                        "processor": processor,
                        "exception": err,
                        **self.log_context,
                    },
                )

    @property
    def hashes(self) -> Dict[str, set[str]]:
        """Return hashes for the most recently seen log entries.

        :return: A dictionary of log entry hashes, keyed by their pointer.
        """
        default: Dict[str, set[str]] = {}
        if self._hashes:
            return self._hashes

        try:
            self._hashes[self.pointer] = set(
                json.loads(
                    self._cache.get(
                        self.cache_key(CACHE_KEY_SEEN),
                        self.operation,
                    )
                )
            )
        except (TypeError, json.decoder.JSONDecodeError) as err:
            self.logger.warning(
                "Deduplication hashes in the cache appear to be malformed, ignoring.",
                extra={"exception": err, **self.log_context},
            )
            return default
        except NotFoundException:
            return default

        return self._hashes

    @hashes.setter
    def hashes(self, value: Dict[str, set[str]]):
        """Sets recent log entry hashes in memory.

        :param value: A dictionary of sets to save.
        """
        self._hashes = value

    def save_hashes(self):
        """Saves the log entry hashes to cache."""
        serialized = json.dumps(
            list(self._hashes.get(self.pointer, set())), separators=(",", ":")
        )

        self._cache.set(self.cache_key(CACHE_KEY_SEEN), self.operation, serialized)

    @property
    def pointer_previous(self) -> str:
        """Return the previous pointer, fetching from cache if needed.

        :return: The previous pointer, which will be an empty string if no previous
            pointer was found.
        """
        if self._pointer_previous:
            return self._pointer_previous

        try:
            self._pointer_previous = self._cache.get(
                self.cache_key(CACHE_KEY_POINTER_PREV), self.operation
            )
        except NotFoundException:
            self._pointer_previous = str()

        return self._pointer_previous

    @property
    def pointer_next(self) -> str:
        """Return the currently known next pointer, fetching from cache if needed.

        :return: The next pointer associated with this configured connector.
        """
        if self._pointer_next:
            return self._pointer_next

        # Intentionally allow exceptions to bubble up so the caller can catch if the
        # value is not in the cache.
        self._pointer_next = self._cache.get(
            self.cache_key(CACHE_KEY_POINTER_NEXT), self.operation
        )

        return self._pointer_next

    @pointer_next.setter
    def pointer_next(self, value: str):
        """Sets and saves the next pointer in cache.

        :param value: The value to save as the next pointer.
        """
        self._cache.set(
            self.cache_key(CACHE_KEY_POINTER_NEXT), self.operation, str(value)
        )

        # Always swap the in-memory value last.
        self._pointer_next = str(value)

    @property
    def pointer(self) -> str:
        """Return the currently known pointer, fetching from cache if needed.

        :return: Pointer associated with this configured connector.
        """
        if self._pointer:
            return self._pointer

        # Intentionally allow exceptions to bubble up so the caller can catch if the
        # value is not in the cache.
        self._pointer = self._cache.get(
            self.cache_key(CACHE_KEY_POINTER), self.operation
        )

        return self._pointer

    @pointer.setter
    def pointer(self, value: str):
        """Sets and saves the pointer and the previous pointer to cache.

        :param value: The value to save as the pointer.
        """
        new_value = str(value)
        old_value = self._pointer

        self._cache.set(
            self.cache_key(CACHE_KEY_POINTER),
            self.operation,
            new_value,
        )
        self._cache.set(
            self.cache_key(CACHE_KEY_POINTER_PREV),
            self.operation,
            old_value,
        )

        # Always swap the in-memory value last.
        self._pointer = new_value
        self._pointer_previous = old_value

    @property
    def window_start(self) -> str:
        """Return the window start location from cache, if set.

        :return: The window start location, or an empty string if not found.
        """
        if self._window_start:
            return self._window_start

        try:
            self._window_start = self._cache.get(
                self.cache_key(CACHE_KEY_WINDOW_START), self.operation
            )
        except NotFoundException:
            return str()

        return self._window_start

    @window_start.setter
    def window_start(self, value: str):
        """Sets the window start location in memory.

        :param value: The value to save as the window start.
        """
        self._window_start = str(value)

    def save_window_start(self):
        """Saves the window start location to cache."""
        self._cache.set(
            self.cache_key(CACHE_KEY_WINDOW_START), self.operation, self.window_start
        )

    @property
    def window_end(self) -> Optional[str]:
        """Return the window end location from cache, if set.

        :return: The window end location.
        """
        if self._window_end:
            return self._window_end

        try:
            self._window_end = self._cache.get(
                self.cache_key(CACHE_KEY_WINDOW_END), self.operation
            )
        except NotFoundException:
            return None

        return self._window_end

    @window_end.setter
    def window_end(self, value: str):
        """Sets the window end location in memory.

        :param value: The value to save as the window end.
        """
        self._window_end = str(value)

    def save_window_end(self):
        """Saves the window end location to cache."""
        self._cache.set(
            self.cache_key(CACHE_KEY_WINDOW_END), self.operation, self.window_end
        )

    def lock(self):
        """Attempts to acquire an execution lock for the current connector.

        This will raise a ConcurrencyException if a a valid lock is already present.

        :raises ConcurrencyException: A valid lock is already held, likely the result
            of a concurrent execution of Grove.
        """
        now = datetime.datetime.utcnow()
        expiry = now + datetime.timedelta(seconds=self._lock_duration)

        # If we don't have a lock, acquire one.
        current = self._lock_expiry

        if current is None:
            try:
                current = datetime.datetime.strptime(
                    self._cache.get(self.cache_key(CACHE_KEY_LOCK), self.operation),
                    LOCK_DATE_FORMAT,
                )
            except NotFoundException:
                pass

            # Someone else has the lock.
            if current is not None and current >= now:
                raise ConcurrencyException(
                    f"Valid lock already held and does not expire until {current}"
                )

        # No lock, or lock expired lock? Let's grab one for ourselves. If we own the
        # lock, we'll also re-lock with a new expiry to keep it.
        #
        # We're constraining the set here to the value of our existing lock, the value
        # we just got from the cache, or "not set" (None). This is to try and protect
        # against a TOCTTOU condition. That said, this code can be improved.
        not_set = True
        constraint = None

        if current is not None:
            not_set = False
            constraint = current.strftime(LOCK_DATE_FORMAT)

        try:
            self._cache.set(
                self.cache_key(CACHE_KEY_LOCK),
                self.operation,
                expiry.strftime(LOCK_DATE_FORMAT),
                constraint=constraint,
                not_set=not_set,
            )
        except DataFormatException:
            raise ConcurrencyException(
                "Could not acquire lock, a valid lock already exists"
            )

        # Lock is ours, track it.
        self._lock_expiry = expiry

    def unlock(self):
        """Releases an execution lock for the current connector.

        :raises AccessException: An unexpected error occurred while releasing the lock.
        :raises ConcurrencyException: The lock does not match the expected value. This
            may indicate a concurrent execution of Grove has since taken the lock.
        """
        # If we don't have a lock, do nothing.
        if self._lock_expiry is None:
            return

        # If there's no lock set in the cache, do nothing.
        try:
            current = datetime.datetime.strptime(
                self._cache.get(self.cache_key(CACHE_KEY_LOCK), self.operation),
                LOCK_DATE_FORMAT,
            )
        except NotFoundException:
            return

        # Check if the lock matches what we expect.
        if current != self._lock_expiry:
            raise ConcurrencyException(
                "The cached lock does not match ours. Someone else may have taken the "
                f"lock! Abandoning attempt to unlock ({current} vs {self._lock_expiry})"
            )

        # Delete the lock, with a constraint to ensure it hasn't changed hands.
        constraint = current.strftime(LOCK_DATE_FORMAT)

        try:
            self._cache.delete(
                self.cache_key(CACHE_KEY_LOCK),
                self.operation,
                constraint=constraint,
            )
        except DataFormatException:
            raise ConcurrencyException(
                "Could not delete lock, the lock no longer appears to be ours."
            )

        # Bye-bye lock.
        self._lock_expiry = None
        self._lock_expiry = None



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove local process entrypoint."""

import os
import socket
from typing import Dict

from grove.entrypoints import base


def runtime_information() -> Dict[str, str]:
    """Attempts to determine the runtime, returning the relevant runtime data.

    :return: A dictionary of runtime data.
    """
    # If Nomad, grab the relevant information.
    if os.environ.get("NOMAD_ALLOC_ID", None):
        return {
            "runtime_id": os.environ.get("NOMAD_ALLOC_ID", "NOT_FOUND"),
            "runtime_region": os.environ.get("NOMAD_REGION", "NOT_FOUND"),
            "runtime_job_name": os.environ.get("NOMAD_JOB_NAME", "NOT_FOUND"),
        }

    # If nothing else matched, assume a local process.
    return {
        "runtime_id": str(os.getpid()),
        "runtime_host": socket.gethostname(),
    }


def entrypoint():
    """Grove local process entrypoint."""
    base.entrypoint(context={"runtime": __file__, **runtime_information()})


# Support local development if called as a script.
if __name__ == "__main__":
    entrypoint()



"""{{ cookiecutter.project_description }}."""

from typing import Optional

from pydantic import Field

from grove.outputs import BaseOutput


class Handler(BaseOutput):
    """A Grove handler for {{ cookiecutter.provider_name }} {{ cookiecutter.provider_product }}."""

    class Configuration(BaseOutput.Configuration):
        """Defines environment variables used to configure the {{ cookiecutter.provider_name }} {{ cookiecutter.provider_product }} handler.

        This should also include any appropriate default values for fields which are not
        required.
        """

        required_setting: str = Field(
            default="thing",
            description="A required value (if unset a default will be used).",
        )
        optional_setting: Optional[str] = Field(
            description="An optional configuration value."
        )

        class Config:
            """Allow environment variable override of configuration fields.

            This also enforce a prefix for all environment variables for this handler.
            As an example the field `required_setting` would be set using the
            environment `{{ cookiecutter.project_slug.upper() }}_REQUIRED_SETTING`.
            """

            env_prefix = "{{ cookiecutter.project_slug.upper() }}_"
            case_insensitive = True

    def setup(self):
        """Set up access to output backend.

        :raises AccessException: There was an issue ...
        """
        pass

    def submit(
        self,
        data: bytes,
        connector: str,
        identity: str,
        operation: str,
        part: int = 0,
        kind: Optional[str] = ".json.gz",
        descriptor: Optional[str] = "logs",
    ):
        """Persists captured data to ...

        :param data: Log data to write.
        :param connector: Name of the connector which retrieved the data.
        :param identity: Identity the collected data was collect for.
        :param operation: Operation the collected logs are associated with.
        :param part: Number indicating which part of the same log stream this file
            contains data for. This is used to indicate that the logs are from the same
            collection, but have been broken into smaller files for downstream
            processing.
        :param kind: An optional file suffix to use for files written.
        :param descriptor: An optional descriptor of the log stream being written.

        :raises AccessException: An issue occurred when writing data.
        """
        pass



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements integration tests for the GSuite Activities collector."""

import os
import unittest
from unittest.mock import patch

from googleapiclient.http import HttpMockSequence

from grove.connectors.gsuite.activities import Connector
from grove.models import ConnectorConfig
from tests import mocks


class MockSequence(HttpMockSequence):
    def close(self, *args, **kwargs):
        pass


class GSuiteActivitiesTestCase(unittest.TestCase):
    """Implements integration tests for the GSuite Activities collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="1FEEDFEED1",
                key="{}",
                name="test",
                connector="test",
                operation="admin",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @patch("google.oauth2.service_account.Credentials.from_service_account_info")
    @patch("googleapiclient.discovery.build")
    @patch("grove.connectors.gsuite.activities.Connector.get_http_transport")
    def test_collect_pagination(self, mock_transport, mock_request, mock_auth):
        """Ensure collection works as expected."""
        mock_transport.return_value = MockSequence(
            [
                (
                    {"status": "200"},
                    open(
                        os.path.join(self.dir, "fixtures/gsuite/activities/001.json"),
                        "r",
                    ).read(),
                ),
                (
                    {"status": "200"},
                    open(
                        os.path.join(self.dir, "fixtures/gsuite/activities/002.json"),
                        "r",
                    ).read(),
                ),
            ],
        )
        self.connector.run()

        self.assertEqual(self.connector._saved["logs"], 2)
        self.assertEqual(self.connector.pointer, "2021-10-27T23:59:31.657Z")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""1Password Audit API client."""

import logging
import time
from typing import Any, Dict, Optional

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_HOSTNAME = "events.1password.com"
API_PAGE_SIZE = 1000


class Client:
    def __init__(
        self,
        hostname: Optional[str] = API_HOSTNAME,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new 1Password audit log client.

        :param hostname: Hostname of the 1Password API to interact with.
        :param token: 1Password token to authenticate with.
        :param retry: Automatically retry if recoverable errors are encountered, such as
            rate-limiting.
        """
        self.retry = retry
        self.hostname = hostname
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

    def _post(
        self,
        url: str,
        payload: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A POST wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP POST against.
        :param payload: Dictionary of data to pass as JSON in the request.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.post(
                    url,
                    json=payload,
                    headers=self.headers,
                    params=params,
                )
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # Retry on rate-limit, but only if requested.
                if getattr(err.response, "status_code", None) == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(1)
                        continue
                    else:
                        raise RateLimitException(err)

                raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_events(
        self,
        event_type: str,
        cursor: Optional[str] = None,
        start_time: Optional[str] = None,
    ) -> AuditLogEntries:
        """Returns a list of logs from a specified endpoint.

        :param event_type: The API endpoint name and type of logs we're pulling.
        :param cursor: Cursor to use when fetching results. Supersedes other parameters.
        :param start_time: The ISO Format timestamp to query logs since.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        url = f"https://{self.hostname}/api/v1/{event_type}"

        # Use the cursor URL if set, otherwise construct the initial query.
        if cursor is not None:
            self.logger.debug(
                "Collecting next page with provided cursor", extra={"cursor": cursor}
            )
            result = self._post(url, payload={"cursor": cursor})
        else:
            # See psf/requests issue #2651 for why we can happily pass in None values
            # and not have the request key added to the URI.
            result = self._post(
                url,
                payload={"limit": API_PAGE_SIZE, "start_time": start_time},
            )

        # Check if pagination is required.
        if result.body.get("has_more", False):
            cursor = result.body.get("cursor")
        else:
            cursor = None

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body.get("items", []))

    def get_signinattempts(
        self,
        cursor: Optional[str] = None,
        start_time: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of signing attempt logs.

        :param cursor: Cursor to use when fetching results. Supersedes other parameters.
        :param start_time: The ISO Format timestamp to query logs since.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_events(
            event_type="signinattempts", cursor=cursor, start_time=start_time
        )

    def get_itemusages(
        self,
        cursor: Optional[str] = None,
        start_time: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of modified, accessed, or used items from a shared vault.

        :param cursor: Cursor to use when fetching results. Supersedes other parameters.
        :param start_time: The ISO Format timestamp to query logs since.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_events(
            event_type="itemusages", cursor=cursor, start_time=start_time
        )

    def get_auditevents(
        self,
        cursor: Optional[str] = None,
        start_time: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of actions performed by members of a 1Password account.

        :param cursor: Cursor to use when fetching results. Supersedes other parameters.
        :param start_time: The ISO Format timestamp to query logs since.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_events(
            event_type="auditevents", cursor=cursor, start_time=start_time
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""A framework for collecting and transforming SaaS logs."""

from grove import (
    caches,  # noqa: F401
    configs,  # noqa: F401
    connectors,  # noqa: F401
    constants,  # noqa: F401
    entrypoints,  # noqa: F401
    exceptions,  # noqa: F401
    helpers,  # noqa: F401
    logging,  # noqa: F401
    models,  # noqa: F401
    outputs,  # noqa: F401
    processors,  # noqa: F401
    types,  # noqa: F401
)
from grove.__about__ import *  # noqa: F401, F403



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for connector event deduplication."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove import constants
from grove.connectors import ConnectorConfig
from tests import mocks
from tests.connectors.test import Connector


class ConnectorDeduplicationTestCase(unittest.TestCase):
    """Implements unit tests for connector event deduplication."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

    def add_response(self, path: str = None, status: int = 200):
        """Adds a response to responses."""
        body = None
        if path:
            body = bytes(open(path, "r").read(), "utf-8")

        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=status,
            content_type="application/json",
            body=body,
        )

    @responses.activate
    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_deduplication_chronological(self):
        """Ensure deduplication works as expected for chronological data."""
        config = ConnectorConfig(
            identity="ORGANISATION",
            key="SOMETHING_SECURE",
            name="UNIT_TEST",
            connector="test_only_connector",
        )
        context = {
            "runtime": "test_harness",
            "runtime_id": "NA",
        }
        first_collection = Connector(config=config, context=context)

        # Load all simulated responses in order.
        for file in ["001.json", "002.json", "003.json"]:
            self.add_response(
                path=os.path.join(self.dir, f"fixtures/grove/chronological/{file}")
            )

        # Run a full collection first and ensure all is as we expect.
        first_collection.run()
        self.assertEqual(first_collection._saved["logs"], 7)
        self.assertEqual(first_collection.pointer, "7")

        # Perform a collection with the latest pointer, and ensure no new records are
        # saved - as they're duplicates. This simulates a "return logs inclusive of the
        # current pointer" operation.
        second_collection = Connector(config=config, context=context)

        # In-memory cache does not support locking as it's only intended for local
        # "one-shot" execution, and development use. As a result, we have to currently
        # alias one to the other to simulate this.
        #
        # TODO: Remove the need for this, as it's going to cause confusion in future.
        second_collection._cache._data = first_collection._cache._data

        self.add_response(
            path=os.path.join(self.dir, "fixtures/grove/chronological/003.json")
        )

        second_collection.run()
        self.assertEqual(second_collection._saved["logs"], 0)
        self.assertEqual(second_collection.pointer, "7")

    @responses.activate
    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_deduplication_reverse_chronological(self):
        """Ensure deduplication works as expected for reverse chronological data."""
        config = ConnectorConfig(
            identity="ORGANISATION",
            key="SOMETHING_SECURE",
            name="UNIT_TEST",
            connector="test_only_connector",
        )
        context = {
            "runtime": "test_harness",
            "runtime_id": "NA",
        }

        # Hot patch the connector to work in reverse chronological order.
        first_collection = Connector(config=config, context=context)

        # This is very naughty.
        first_collection.LOG_ORDER = constants.REVERSE_CHRONOLOGICAL

        # Load all simulated responses in order.
        for file in ["001.json", "002.json"]:
            self.add_response(
                path=os.path.join(
                    self.dir, f"fixtures/grove/reverse_chronological/{file}"
                )
            )

        # Run a full collection first and ensure all is as we expect.
        first_collection.run()
        self.assertEqual(first_collection._saved["logs"], 7)
        self.assertEqual(first_collection.pointer, "7")

        # Perform a collection with the latest pointer, and ensure that only new records
        # are returned.
        second_collection = Connector(config=config, context=context)

        # In-memory cache does not support locking as it's only intended for local
        # "one-shot" execution, and development use. As a result, we have to currently
        # alias the state of one cache to the other to simulate this.
        #
        # TODO: Remove the need for this, as it's going to cause confusion in future.
        second_collection._cache._data = first_collection._cache._data

        self.add_response(
            path=os.path.join(self.dir, "fixtures/grove/reverse_chronological/003.json")
        )

        second_collection.run()
        self.assertEqual(second_collection._saved["logs"], 1)
        self.assertEqual(second_collection.pointer, "7")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Exceptions used by Grove."""


class GroveException(Exception):
    """All exceptions should inherit from this to allow for hierarchical handling."""


class ConfigurationException(GroveException):
    """Indicates that a configuration related error has occurred."""


class ConnectorMissingException(GroveException):
    """Indicates that a requested connector was not found."""


class ConcurrencyException(GroveException):
    """Indicates that Grove may be running in another location concurrently."""


class NotImplementedException(GroveException):
    """Indicates that the requested functionality has not been implemented."""


class NotFoundException(GroveException):
    """Indicates that the requested entity was not found."""


class RequestFailedException(GroveException):
    """Indicates that an upstream request failed for an unhandled reason."""


class RateLimitException(GroveException):
    """Indicate that an upstream rate-limit was encountered."""


class AccessException(GroveException):
    """Indicates an issue occurred while attempting to access the requested resource."""


class DataFormatException(GroveException):
    """Indicates an issue occurred while attempting to process data."""


class ProcessorError(GroveException):
    """Indicates that an error occurred when setting up or calling a processor."""



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove AWS S3 output handler."""

import datetime
import os
from typing import Optional

from boto3.session import Session
from botocore.exceptions import BotoCoreError, ClientError
from pydantic import Field

from grove.constants import DATESTAMP_FORMAT
from grove.exceptions import AccessException
from grove.outputs import BaseOutput

OBJECT_KEY = (
    "{descriptor}{connector}/{identity}/{year}/{month}/{day}/"
    "{operation}-{datestamp}.{part}{kind}"
)


class Handler(BaseOutput):
    """This output handler allows Grove to write collected logs to an AWS S3 bucket."""

    class Configuration(BaseOutput.Configuration):
        """Defines environment variables used to configure the AWS S3 handler.

        This should also include any appropriate default values for fields which are not
        required.
        """

        bucket: str = Field(
            description="The name of the S3 bucket to output logs to.",
        )
        aws_access_key_id: Optional[str] = Field(
            description="An optional AWS access key to use when authenticating",
            default=os.environ.get("AWS_ACCESS_KEY_ID"),
        )
        aws_secret_access_key: Optional[str] = Field(
            description="An optional AWS secret key to use when authenticating",
            default=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        )
        aws_session_token: Optional[str] = Field(
            description="An optional AWS session token to use when authenticating",
            default=os.environ.get("AWS_SESSION_TOKEN"),
        )
        assume_role_arn: Optional[str] = Field(
            description="An optional AWS role to assume when authenticating with AWS.",
            default=None,
        )
        bucket_region: Optional[str] = Field(
            description="The region that S3 the bucket exists in (default us-east-1)",
            default=os.environ.get("AWS_REGION", "us-east-1"),
        )

        class Config:
            """Allow environment variable override of configuration fields.

            This also enforce a prefix for all environment variables for this handler.
            As an example the field `bucket` would be set using the environment variable
            `GROVE_OUTPUT_AWS_S3_BUCKET`.
            """

            env_prefix = "GROVE_OUTPUT_AWS_S3_"
            case_insensitive = True

    def setup(self):
        """Sets up access to S3.

        This handler also attempt to assume a configured role in order to allow
        cross-account use - if required.

        :raises ConfigurationException: There was an issue with configuration.
        :raises AccessException: An issue occurred when accessing S3.
        """
        # Explicit calls to session are mostly used to allow mocks during testing.
        session = Session()

        # Only add in optional arguments if configured.
        client_kwargs = {}

        if self.config.aws_access_key_id:
            client_kwargs["aws_access_key_id"] = self.config.aws_access_key_id
            client_kwargs["aws_secret_access_key"] = self.config.aws_secret_access_key

        if self.config.aws_session_token:
            client_kwargs["aws_session_token"] = self.config.aws_session_token

        # If a role was specified, ensure we assume it and use STS tokens to interact
        # with S3.
        try:
            if not self.config.assume_role_arn:
                self.s3 = session.client(
                    "s3",
                    region_name=self.config.bucket_region,
                    **client_kwargs,
                )
            else:
                sts = session.client(
                    "sts",
                    region_name=self.config.bucket_region,
                    **client_kwargs,
                )
                role = sts.assume_role(
                    RoleArn=self.config.assume_role_arn,
                    RoleSessionName="GroveOutputWriter",
                )
                self.s3 = session.client(
                    "s3",
                    region_name=self.config.bucket_region,
                    aws_access_key_id=role["Credentials"]["AccessKeyId"],
                    aws_secret_access_key=role["Credentials"]["SecretAccessKey"],
                    aws_session_token=role["Credentials"]["SessionToken"],
                )
        except (ClientError, BotoCoreError) as err:
            raise AccessException(f"Output handler is unable to access AWS S3: {err}")

    def submit(
        self,
        data: bytes,
        connector: str,
        identity: str,
        operation: str,
        part: int = 0,
        kind: Optional[str] = ".json.gz",
        descriptor: Optional[str] = "logs/",
    ):
        """Persists captured data to an S3 compatible object store.

        :param data: Log data to write to S3.
        :param connector: Name of the connector which retrieved the data.
        :param identity: Identity the collected data was collect for.
        :param operation: Operation the collected logs are associated with.
        :param part: Number indicating which part of the same log stream this file
            contains data for. This is used to indicate that the logs are from the same
            collection, but have been broken into smaller files for downstream
            processing.
        :param kind: An optional file suffix to use for objects written.
        :param descriptor: An optional path to append to the beginning of the output
            S3 key.

        :raises AccessException: An issue occurred when accessing S3.
        """
        # Append a trailing slash to the descriptor if set - to form a path.
        if descriptor and not descriptor.endswith("/"):
            descriptor = f"{descriptor}/"

        try:
            datestamp = datetime.datetime.utcnow()
            self.s3.put_object(
                Body=data,
                Bucket=self.config.bucket,
                Key=OBJECT_KEY.format(
                    part=part,
                    connector=connector,
                    identity=identity,
                    operation=operation,
                    year=datestamp.strftime("%Y"),
                    month=datestamp.strftime("%m"),
                    day=datestamp.strftime("%d"),
                    datestamp=datestamp.strftime(DATESTAMP_FORMAT),
                    descriptor=descriptor,
                    kind=kind,
                ),
            )
        except ClientError as err:
            raise AccessException(f"Unable to write object to AWS S3: {err}")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove local file configuration handler."""

import glob
import json
import logging
from json import JSONDecodeError
from typing import List

from pydantic import BaseSettings, Field, ValidationError

from grove.configs import BaseConfig
from grove.exceptions import ConfigurationException, DataFormatException
from grove.helpers import parsing
from grove.models import ConnectorConfig


class Configuration(BaseSettings):
    """Defines environment variables used to configure the local file handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    path: str = Field(
        description="The directory path containing connector configuration documents.",
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `path` would be set using the environment variable
        `GROVE_CONFIG_LOCAL_FILE_PATH`.
        """

        env_prefix = "GROVE_CONFIG_LOCAL_FILE_"
        case_insensitive = True


class Handler(BaseConfig):
    """A configuration handler to read configuration documents from local files."""

    def __init__(self):
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()  # type: ignore
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

    def get(self, id: str = "") -> List[ConnectorConfig]:
        """Get and return one or more connector configuration objects from local files.

        :param id: Not used.

        :return: A list of connector configuration documents (JSON) as strings.
        """
        connectors = []

        # Generate a list of documents for later processing.
        for path in glob.glob(f"{self.config.path}/**/*.json", recursive=True):
            with open(path, "r") as f:
                # We don't want a single bad connector configuration document to break
                # collection, so log an error and continue on a bad document.
                try:
                    connectors.append(ConnectorConfig(**json.load(f)))
                except (JSONDecodeError, ValidationError, DataFormatException) as err:
                    self.logger.error(
                        "Unable to load connector configuration",
                        extra={"path": path, "exception": err},
                    )
                    continue

        return connectors



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Tines Audit API client.

This is a bare-bones client designed to interact with audit related APIs only.
"""

import logging
import time
from typing import Dict, Optional

import jmespath
import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://{identity}.{domain}/api/v1"
API_PAGE_SIZE = 500


class Client:
    def __init__(
        self,
        identity: Optional[str] = None,
        domain: str = "tines.com",
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new Tines API client.

        :param token: Tines API token.
        :param identity: The name of the Tines tenant to collect logs from.
        :param domain: The Tines domain to use when constructing API URLs. This is not
            usually required if using Tines hosted tenants.
        :param retry: Automatically retry if recoverable errors are encountered, such as
            rate-limiting.
        """
        self.retry = retry
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "content-type": "application/json",
            "x-user-token": token,
        }
        self._api_base_uri = API_BASE_URI.format(identity=identity, domain=domain)

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(
                    url,
                    headers=self.headers,  # type: ignore
                    params=params,
                )
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # Retry on rate-limit, but only if requested.
                if getattr(err.response, "status_code", None) == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(int(err.response.headers.get("Retry-After", "1")))
                        continue
                    else:
                        raise RateLimitException(err)

                raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def list_audit_logs(
        self,
        after: Optional[str] = None,
        operation_name: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs which match the provided filters.

        :param after: An RFC3339 timestamp, without milliseconds, to collect logs after.
        :param operation_name: An optional operation to collect logs for.
        :param cursor: Cursor to use when fetching events (pagination). This is the
            page number in the context of Tines.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # See psf/requests issue #2651 for why we can happily pass in None values and
        # not have the request key added to the URI.
        result = self._get(
            f"{self._api_base_uri}/audit_logs",
            params={
                "page": cursor,
                "after": after,
                "operation_name": operation_name,
                "per_page": str(API_PAGE_SIZE),
            },
        )

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(
            cursor=jmespath.search("meta.next_page", result.body),
            entries=jmespath.search("audit_logs", result.body),
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides Grove secret storage using supported backends."""

import abc
import logging
from typing import List

from grove.exceptions import AccessException, DataFormatException
from grove.models import ConnectorConfig, decode


class BaseSecret(abc.ABC):
    def __init__(self):
        """Provides the basis for all Grove secret backends."""
        self.logger = logging.getLogger(__name__)

    @abc.abstractmethod
    def get(self, path: str) -> str:
        """Gets the secret with the given identifier from the given backend.

        :param path: The path to the credential to get.
        :return: The decoded plain-text credential for use by connectors.
        """
        pass

    def load(self, configurations: List[ConnectorConfig]) -> List[ConnectorConfig]:
        """Gets secrets from the backend, inserting them into configuration objects.

        This method should not be implemented by secrets handlers, as the operations
        should be identical between implementations (calls to get()).

        :param configurations: A list of ConnectorConfig objects from the configuration
            backend.

        :return: A list of ConnectorConfig objects with secrets included.
        """
        ready = []

        for configuration in configurations:
            # Fetch the the real secret from the backend using the identifier from the
            # 'secrets' object - decoding it if required.
            try:
                for field, identifier in configuration.secrets.items():
                    self.logger.debug(
                        "Attempting to get query secret from backend",
                        extra={
                            "field": field,
                            "identifier": identifier,
                            "document": configuration.name,
                        },
                    )
                    candidate = self.get(identifier)

                    # Decode the value, if required.
                    if field in configuration.encoding:
                        candidate = decode(candidate, configuration.encoding[field])

                    setattr(configuration, field, candidate)
            except DataFormatException as err:
                self.logger.error(
                    "Unable to decode secret for connector, skipping",
                    extra={
                        "document": configuration.name,
                        "field": field,
                        "exception": err,
                    },
                )
                continue
            except (AccessException, IndexError) as err:
                self.logger.error(
                    "Unable to get secret for connector, skipping",
                    extra={
                        "document": configuration.name,
                        "field": field,
                        "exception": err,
                    },
                )
                continue

            # Some connectors may not have any secrets.
            ready.append(configuration)

        return ready



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements integration tests for the TFC Audit collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove.connectors.tfc.audit_trails import Connector
from grove.models import ConnectorConfig
from tests import mocks


class TFCAuditTestCase(unittest.TestCase):
    """Implements integration tests for the TFC Audit collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="1FEEDFEED1",
                key="token",
                name="test",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_collect_rate_limit(self):
        """Ensure rate-limit retires are working as expected."""
        # Rate limit the first request.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=429,
            content_type="application/json",
            body=bytes(),
        )

        # Succeed on the second.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/tfc/audit_trails/002.json"), "r"
                ).read(),
                "utf-8",
            ),
        )

        # Ensure time.sleep is called with the correct value in response to a
        # rate-limit.
        with patch("time.sleep", return_value=None) as mock_sleep:
            self.connector.run()
            mock_sleep.assert_called_with(1)

    @responses.activate
    def test_collect_pagination(self):
        """Ensure pagination is working as expected."""
        # Succeed with a cursor returned (to indicate paging is required).
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/tfc/audit_trails/001.json"), "r"
                ).read(),
                "utf-8",
            ),
        )

        # The last "page" returns an empty cursor.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/tfc/audit_trails/002.json"), "r"
                ).read(),
                "utf-8",
            ),
        )

        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 3)
        self.assertEqual(self.connector.pointer, "2020-06-30T17:52:46.000Z")

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/tfc/audit_trails/003.json"), "r"
                ).read(),
                "utf-8",
            ),
        )

        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 5)
        self.assertEqual(self.connector.pointer, "2020-06-30T17:52:46.000Z")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Zoom client."""

import base64
import os
import re
import unittest
from unittest.mock import patch

import responses
from responses import matchers

from grove.connectors.zoom import api
from tests import mocks


class ZoomClientTestCases(unittest.TestCase):
    """Implements unit tests for the Zoom client."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.identity = "ACCOUNT_ID"
        self.key = "CLIENT_SECRET"
        self.client = None

    @responses.activate
    def test_get_bearer_token(self):
        """Ensure we get bearer token from client-id client-secret."""
        set_expected_basic_auth = str(
            base64.b64encode(bytes(f"{self.identity}:{self.key}", "utf-8")),
            "utf-8",
        )
        # auth response
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/zoom/client/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        # token hardcoded from fixtures file
        resp = {"access_token": "testtoken"}
        expected = "testtoken"
        self.assertIsNone(self.client)
        self.client = api.Client(identity=self.identity, key=self.key)
        self.assertEqual(resp.get("access_token"), expected)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements integration tests for the SalesForce Event Log collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove.connectors.sf.event_log import Connector
from grove.exceptions import RequestFailedException
from grove.models import ConnectorConfig
from tests import mocks


class SFEventLogTestCase(unittest.TestCase):
    """Implements integration tests for the SalesForce Evennt Log collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="Someuser",
                key="token",
                name="test",
                connector="test",
                token="12345",
                operation="Login",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_collect_fail_on_server_error(self):
        """Ensure server errors raise an appropriate exception."""
        # Ensure authentication succeeds (POST to SF).
        login_response = bytes(
            open(os.path.join(self.dir, "fixtures/sf/event_log/login.xml"), "r").read(),
            "utf-8",
        )
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            body=login_response,
            content_type="application/xml",
        )

        # Ensure first query returns an HTTP 500 (GET to SF)
        query_response = bytes(
            open(os.path.join(self.dir, "fixtures/sf/event_log/error.xml"), "r").read(),
            "utf-8",
        )
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=500,
            body=query_response,
            content_type="application/xml",
        )

        with self.assertRaises(RequestFailedException):
            self.connector.collect()

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        # Ensure authentication succeeds (POST to SF).
        login_response = bytes(
            open(os.path.join(self.dir, "fixtures/sf/event_log/login.xml"), "r").read(),
            "utf-8",
        )
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            body=login_response,
            content_type="application/xml",
        )

        # Ensure EventLogFile query returns a 200 (GET to SF).
        query_response = bytes(
            open(os.path.join(self.dir, "fixtures/sf/event_log/001.json"), "r").read(),
            "utf-8",
        )
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            body=query_response,
            content_type="application/xml",
        )

        # Ensure LogFile query returns a 200 (GET to SF).
        log_response = bytes(
            open(os.path.join(self.dir, "fixtures/sf/event_log/001.csv"), "r").read(),
            "utf-8",
        )
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            body=log_response,
            content_type="text/csv",
        )

        # Check the pointer matches the latest value, and that the expected number of
        # logs were returned.
        self.connector.collect()
        self.assertEqual(self.connector._saved["logs"], 2)
        self.assertEqual(self.connector.pointer, "2038-01-19T03:00:00.000Z")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Slack Audit API client.

The official Slack SDK does not currently support the Audit API, this client has been
created in the interim.
"""

import logging
import time
from typing import Dict, Optional

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://api.slack.com/audit/v1"


class Client:
    def __init__(
        self,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new Slack audit API client.

        :param token: Slack API Bearer token.
        :param retry: Automatically retry if recoverable errors are encountered, such as
            rate-limiting.
        """
        self.retry = retry
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {token}",
        }

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # Retry on rate-limit, but only if requested.
                if getattr(err.response, "status_code", None) == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(int(err.response.headers.get("Retry-After", "1")))
                        continue
                    else:
                        raise RateLimitException(err)

                raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_logs(
        self,
        latest: Optional[str] = None,
        oldest: Optional[str] = None,
        action: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs which match the provided filters.

        :param latest: Unix timestamp of the most recent event to include (inclusive).
        :param oldest: Unix timestamp of the least recent event to include (inclusive).
        :param action: Name of the action to request events for.
        :param cursor: Cursor to use when fetching events (pagination).

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # See psf/requests issue #2651 for why we can happily pass in None values and
        # not have the request key added to the URI.
        result = self._get(
            f"{API_BASE_URI}/logs",
            params={
                "latest": latest,
                "oldest": oldest,
                "action": action,
                "cursor": cursor,
                "limit": "1000",
            },
        )

        # Slack appears to return an empty string for a cursor if there isn't one, so
        # swap this for None in this case to avoid having to rely on "falsy" conditions.
        cursor = result.body.get("response_metadata", {}).get("next_cursor", None)
        if cursor == "":
            cursor = None

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body.get("entries", []))



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Tines Audit connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.tines.api import Client
from grove.constants import OPERATION_DEFAULT, REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "tines_audit_logs"
    POINTER_PATH = "created_at"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    @property
    def domain(self):
        """Fetches the Tines domain suffix from the configuration.

        This field is used to allow configuration of collection of log data from
        specific non 'tines.com' instances. Usually, this will not need to be changed,
        as the configured identity (tenant name) will be appended to this domain to form
        the full FQDN.

        If the required tenant is under 'tines.com', only the usual identity field need
        be set.

        :return: The "domain" portion of the connector's configuration.
        """
        try:
            return self.configuration.domain
        except AttributeError:
            return "tines.com"

    def collect(self):
        """Collects all logs from the Tines Audit API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(
            token=self.key,
            domain=self.domain,
            identity=self.identity,
        )
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. The Tines API returns timestamps as RFC3339, and
        # without milliseconds.
        now = datetime.utcnow()

        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (now - timedelta(days=7)).isoformat(
                sep="T",
                timespec="seconds",
            ) + "Z"

        # Set the operation name to collect to 'None' if none is specified - as the
        # Grove default is 'all'.
        operation = None

        if self.operation != OPERATION_DEFAULT:
            operation = self.operation

        # Page over data using the cursor, saving returned data page by page.
        while True:
            log = client.list_audit_logs(
                after=self.pointer,
                cursor=cursor,
                operation_name=operation,
            )

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Google GSuite Alerts connector for Grove."""

import json
from datetime import datetime, timedelta

import google_auth_httplib2
import httplib2
from google.auth.exceptions import GoogleAuthError
from google.oauth2 import service_account
from googleapiclient.discovery import build
from googleapiclient.errors import Error

from grove.connectors import BaseConnector
from grove.constants import CHRONOLOGICAL
from grove.exceptions import (
    ConfigurationException,
    NotFoundException,
    RequestFailedException,
)

# This connector is only interested in the GSuite Alerts API.
SCOPES = ["https://www.googleapis.com/auth/apps.alerts"]


class Connector(BaseConnector):
    NAME = "gsuite_alerts"
    POINTER_PATH = "createTime"
    LOG_ORDER = CHRONOLOGICAL

    def collect(self):
        """Collects all alerts from the Google GSuite Alerts API.

        As the Google APIs use OAuth 2.0 2LO ('two-legged OAuth') which contains a
        number of fields inside of a JSON 'service account file' the key and identity
        are treated a little differently in this connector.

        Rather than the key being a single authentication token, the key should contain
        the entire 'service account file' in JSON format - as generated by the Google
        API console.

        The identity must be the name of a service account which has been granted domain
        wide delegation. Please see the following guides for more information:

          https://developers.google.com/admin-sdk/alertcenter/guides/prerequisites
          https://developers.google.com/admin-sdk/directory/v1/guides/delegation

        :raises RequestFailedException: An HTTP request failed.
        """
        cursor = str()
        http = google_auth_httplib2.AuthorizedHttp(
            self.get_credentials(),
            http=self.get_http_transport(),
        )

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the GSuite audit API the pointer is the
        # value of the "createTime" field from the latest record retrieved from the
        # API.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=7)).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )

        # Page over all alerts since the last pointer.
        more_requests = True

        with build("alertcenter", "v1beta1", http=http) as service:
            while more_requests:
                # The cursor may work with the filter, but let's not combine them, just
                # to try and avoid any strange and difficult to debug behaviour.
                self.logger.debug(
                    "Requesting GSuite alert list.", extra={"operation": self.operation}
                )
                if cursor:
                    self.logger.debug(
                        "Using pageToken as cursor to request next page of results",
                        extra={"cursor": cursor},
                    )
                    request = service.alerts().list(
                        orderBy="createTime asc", pageToken=cursor
                    )
                else:
                    # Google's APIs return timestamps with microseconds, as a result the
                    # potential for two records to occur at the same time and one be
                    # missed is EXTREMELY low. As a result, we'll use > rather than >=
                    # here.
                    self.logger.debug(
                        "Using createTime from pointer",
                        extra={"pointer": self.pointer},
                    )
                    request = service.alerts().list(
                        orderBy="createTime asc",
                        filter=f'createTime > "{self.pointer}"',
                    )

                # Page over results and save.
                try:
                    results = request.execute()
                    alerts = results.get("alerts", [])

                    self.logger.debug(
                        "Got alerts from the GSuite API", extra={"count": len(alerts)}
                    )
                    self.save(alerts)
                except (GoogleAuthError, Error) as err:
                    raise RequestFailedException(f"Request to GSuite API failed: {err}")

                # Determine whether we're still paging.
                if "nextPageToken" not in results:
                    self.logger.debug("No nextPageToken, finishing collection.")
                    more_requests = False
                    break

                self.logger.debug("nextPageToken is present in response, paging.")
                cursor = results["nextPageToken"]
                more_requests = True

    def get_http_transport(self):
        """Creates an HTTP object for use by the Google API Client.

        :return: An httplib2.Http object for use with the Google API client.
        """
        return httplib2.Http()

    def get_credentials(self):
        """Generates and returns a credentials instance from the connector's configured
        service account info. This is used for required to perform operations using the
        Google API client.

        :return: A credentials instance built from configured service account info.

        :raises ConfigurationException: There is an issue with the configuration
            for this connector.
        """
        try:
            service_account_info = json.loads(self.key)
        except json.JSONDecodeError as err:
            raise ConfigurationException(
                f"Unable to load service account JSON for {self.identity}: {err}"
            )

        # Construct the credentials, including scopes and delegation.
        try:
            credentials = service_account.Credentials.from_service_account_info(
                service_account_info,
                scopes=SCOPES,
                subject=self.identity,
            )
        except ValueError as err:
            raise ConfigurationException(
                "Unable to generate credentials from service account info for "
                f"{self.identity}: {err}"
            )

        return credentials



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides Mock implementations for unit and integration tests."""

from typing import Any

from grove.caches import local_memory
from grove.constants import PLUGIN_GROUP_CACHE, PLUGIN_GROUP_OUTPUT
from grove.helpers import plugin
from tests.mocks import output  # noqa: F401


def load_handler(name: str, group: str, *args, **kwargs) -> Any:
    """Wraps handler loading to load predefined mocks for a given group."""
    if group == PLUGIN_GROUP_OUTPUT:
        return output.TestHandler()

    if group == PLUGIN_GROUP_CACHE:
        return local_memory.Handler()

    cls = plugin.lookup_handler(name, group).load()
    return cls(*args, **kwargs)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""GitHub Audit connector for Grove."""

from datetime import datetime, timedelta, timezone

from grove.connectors import BaseConnector
from grove.connectors.github.api import Client
from grove.constants import CHRONOLOGICAL
from grove.exceptions import ConfigurationException, NotFoundException

DATESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Connector(BaseConnector):
    NAME = "github_audit_log"
    LOG_ORDER = CHRONOLOGICAL

    # Double quoting is required so that jmespath understands that @timestamp is the
    # field to be extracted - due to the special character at the start ('@').
    POINTER_PATH = '"@timestamp"'

    @property
    def delay(self):
        """Defines the amount of time to delay collection of logs (in minutes).

        This is used to allow time for logs to become 'consistent' before they are
        collected. This is required as Github backfills log entries but unfortunately do
        not provide any guidance around 'lag' time, or guarantees on availability and
        delivery.

        As a result of these constraints, this value is configurable to allow operators
        to preference consistency over speed of delivery, and vice versa. For example,
        a delay of 20 would instruct Grove to only collect logs after they are at least
        20 minutes old.

        This defaults to 0 (no delay).

        :return: The "delay" component of the connector configuration.
        """
        try:
            candidate = self.configuration.delay
        except AttributeError:
            return 0

        try:
            candidate = int(candidate)
        except ValueError as err:
            raise ConfigurationException(
                f"Configured 'delay' is not valid. Value must be an integer. {err}"
            )

        return candidate

    @property
    def scope(self):
        """Fetches the configured Github scope.

        This is used to control whether the connector should collect logs for a Github
        enterprise, or an organisation. This defaults to "orgs".

        :return: The "scope" component of the connector configuration.
        """
        try:
            candidate = self.configuration.scope
        except AttributeError:
            return "orgs"

        # Check that this style of account is supported.
        SUPPORTED = ["enterprises", "orgs"]

        if candidate.lower() not in SUPPORTED:
            raise ConfigurationException(
                f"Configured 'scope' is not valid. Only {SUPPORTED} are supported."
            )

        return candidate

    @property
    def fqdn(self):
        """Fetches the configured Github API FQDN, or the default (SaaS).

        :return: The "fqdn" component of the connector configuration.
        """
        try:
            return self.configuration.fqdn
        except AttributeError:
            return "api.github.com"

    def collect(self):
        """Collects all logs from the GitHub Audit API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(
            token=self.key,
            scope=self.scope,
            identity=self.identity,
            hostname=self.fqdn,
        )
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the GitHub audit API the pointer is the
        # value of the "created_at" field from the latest record retrieved from the
        # API - which is in milliseconds since epoch format.
        try:
            _ = self.pointer
        except NotFoundException:
            # Precision doesn't matter too much here, as the GitHub API currently
            # doesn't appear to support millisecond granularity in filters.
            self.pointer = str(
                int((datetime.now(timezone.utc) - timedelta(days=7)).timestamp() * 1000)
            )

        # Transform the pointer into an ISO8601 compatible date and construct the search
        # phrase.
        start = datetime.utcfromtimestamp(int(self.pointer) / 1000)
        end = datetime.utcnow() - timedelta(minutes=self.delay)

        # Get log data from the upstream API, paging as required.
        while True:
            if end <= start:
                self.logger.debug(
                    "Collection end time is prior to start, skipping.",
                    extra={
                        "start": start.strftime(DATESTAMP_FORMAT),
                        "end": end.strftime(DATESTAMP_FORMAT),
                    },
                )
                break

            log = client.get_audit_log(
                phrase=(
                    f"created:>={start.strftime(DATESTAMP_FORMAT)} "
                    f"created:<={end.strftime(DATESTAMP_FORMAT)}"
                ),
                include=self.operation,
                cursor=cursor,
            )

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides processors for collected log entries."""

import abc
import logging
from typing import Any, Dict, List

from pydantic import Extra, ValidationError

from grove.exceptions import ConfigurationException
from grove.helpers import parsing
from grove.models import ProcessorConfig


class BaseProcessor(abc.ABC):
    """Provides an abstract base processor which all processors must inherit from."""

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Defines the required configuration and validators for the processor."""

        pass

    def __init__(self, config: Dict[str, Any]):
        """Sets up a Grove processor.

        :param config: The configuration document for this processor, as a dict.
        """
        self.logger = logging.getLogger(__name__)

        # Load and validate configuration. We perform a bit of a strange operation here
        # but our caller needs to have loaded the configuration into a ProcessorConfig
        # already, but we want to re-validate it here. As a result, we convert to a dict
        # and back again.
        try:
            self.configuration = self.Configuration(**config.dict())
        except ValidationError as err:
            raise ConfigurationException(
                f"Processor configuration is invalid. {parsing.validation_error(err)}"
            )

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Performs a set of processes against a log entry.

        :param entry: A collected log entry.

        :returns: The processed log entry in a list. If only a single entry is required
            the list should contain a single element. If the log entry is to be dropped,
            an empty list should be used.
        """
        return [entry]

    def finalize(self):
        """Performs a final set of operations after logs have been saved."""

        return



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Mocks a Grove output provider."""

from grove.outputs import BaseOutput


class TestHandler(BaseOutput):
    __test__ = False

    def submit(self, *args, **kwargs):
        """Does nothing, successfully."""



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the path filtering processor."""

import json
import os
import unittest
from unittest.mock import patch

from grove.models import ProcessorConfig
from grove.processors import filter_paths
from tests import mocks


class ProcessorPathFilterTestCase(unittest.TestCase):
    """Implements unit tests for the path filtering processor."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Setup the processor and associated configuration for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

        # Create a mapping compatible with the Okta system log fixture.
        self.processor = filter_paths.Handler(
            ProcessorConfig(
                name="Filter debugContext",
                processor="filter_paths",
                sources=[
                    "debugContext",
                    "client.geographicalContext",
                ],
            )
        )

    def test_filter_paths(self):
        """Ensure path filtering operates as expected."""
        # Load and process the fixture.
        entries = json.load(
            open(
                os.path.join(self.dir, "fixtures/okta/system_log/001.json"),
                "r",
            )
        )

        # Firstly, ensure the 'debugContext' field exists to begin with.
        self.assertTrue("debugContext" in entries[0])
        self.assertTrue("geographicalContext" in entries[0]["client"])

        # Process a single log entry.
        records = self.processor.process(entries[0])

        # Ensure the 'debugContext' field was removed.
        self.assertFalse("debugContext" in records[0])
        self.assertFalse("geographicalContext" in entries[0]["client"])



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Test connector for Grove."""

import datetime

import requests

from grove.connectors import BaseConnector
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "test_only_connector"
    LOG_ORDER = CHRONOLOGICAL
    POINTER_PATH = "timestamp"

    def collect(self):
        """Collects all logs."""

        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = str(int(datetime.datetime.now().timestamp()))

        while True:
            response = requests.get(
                "http://192.0.2.1",
                timeout=1,
                verify=False,
                params={
                    "start_time": self.pointer,
                },
            )
            entries = response.json().get("logs", [])
            cursor = response.json().get("cursor")

            # Save this batch of log entries.
            self.save(entries)

            # Check if we need to continue paging.
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove local file secrets handler."""

import logging
import os

from pydantic import BaseSettings, Field, ValidationError

from grove.exceptions import AccessException, ConfigurationException
from grove.helpers import parsing
from grove.secrets import BaseSecret


class Configuration(BaseSettings):
    """Defines environment variables used to configure the local file handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    path_prefix: str = Field(
        str(),
        description="An optional prefix to append to configured secret paths.",
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `path` would be set using the environment variable
        `GROVE_SECRET_LOCAL_FILE_PATH_PREFIX`.
        """

        env_prefix = "GROVE_SECRET_LOCAL_FILE_"
        case_insensitive = True


class Handler(BaseSecret):
    """A secret handler to read secrets from local files."""

    def __init__(self):
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()  # type: ignore
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

    def get(self, id: str) -> str:
        """Gets and returns an secret from the specified file path.

        If a path prefix is configured this will be appended to the beginning of the
        configured file path. However, if the path of the secret begins with a '/' it
        the path prefix will be ignored - as it will be considered a fully-qualified
        path specification.

        :param id: The file to read the secret from.

        :return: The plain-text secret, read from the specified file.
        """
        secret = str()
        path = os.path.join(self.config.path_prefix, id)

        try:
            with open(path, "rb") as f:
                secret = str(f.read(), "utf-8").rstrip()
        except (ValidationError, OSError) as err:
            raise AccessException(
                f"Unable to read secret from configured '{path}'. {err}"
            )

        return secret



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

""" Provides post generation hooks to handle optional files.

This is required to ensure that files not relevant for the given project type are
properly removed.
"""

import os
import shutil


def main():
    # Move files around based on the selected project type.
    os.rename(
        "./{{cookiecutter.project_slug}}/inits/{{ cookiecutter.project_type }}.py",
        "./{{cookiecutter.project_slug}}/__init__.py",
    )

    # Remove files only used by specific project types if not required.
    if "{{ cookiecutter.project_type }}" != "connectors":
        os.remove("./{{cookiecutter.project_slug}}/example_logs.py")

    # Remove the handlers directory, as this is only used
    shutil.rmtree("./{{cookiecutter.project_slug}}/inits/")


if __name__ == "__main__":
    main()



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the path extracting processor."""

import json
import os
import unittest
from unittest.mock import patch

from grove.models import ProcessorConfig
from grove.processors import extract_paths
from tests import mocks


class ProcessorPathExtratTestCase(unittest.TestCase):
    """Implements unit tests for the path extracting processor."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Setup the processor and associated configuration for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

        # Create a mapping compatible with the Okta system log fixture.
        self.processor = extract_paths.Handler(
            ProcessorConfig(
                name="ecs",
                processor="extract_paths",
                raw="event.original",
                fields=[
                    {
                        "destination": "@timestamp",
                        "sources": [
                            "published",
                        ],
                    },
                    {
                        "destination": "'source.ip'",
                        "sources": [
                            "client.ipAddress",
                        ],
                    },
                    {
                        "destination": "'ecs.version'",
                        "static": "8.8",
                    },
                    {
                        "destination": "nested.key",
                        "static": "example",
                    },
                    {
                        "destination": "another.nested.key",
                        "sources": [
                            "client.device",
                        ],
                    },
                ],
            )
        )

    def test_extract_paths(self):
        """Ensure path extraction operates as expected."""
        # Load and process the fixture.
        entries = json.load(
            open(os.path.join(self.dir, "fixtures/okta/system_log/001.json"), "r")
        )

        # Process a single target.
        target = entries[0]
        results = self.processor.process(target)

        # Ensure fields are is mapped correctly.
        self.assertEqual(results[0]["source.ip"], "000.000.00.000")
        self.assertEqual(results[0]["@timestamp"], "2021-06-24T00:04:08.123Z")
        self.assertEqual(results[0]["ecs.version"], "8.8")
        self.assertEqual(results[0]["nested"]["key"], "example")
        self.assertEqual(results[0]["another"]["nested"]["key"], "Computer")

        # Ensure the raw message is present.
        self.assertGreater(len(results[0]["event"]["original"]), 0)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Twilio monitor events connector for Grove."""

from datetime import datetime, timedelta, timezone

from twilio.base.exceptions import TwilioException
from twilio.rest import Client

from grove.connectors import BaseConnector
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException, RequestFailedException


class Connector(BaseConnector):
    NAME = "twilio_monitor_events"
    POINTER_PATH = "event_date"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    @property
    def secret(self):
        """Fetches the API secret from the configuration.

        This is required as this is a third authentication element required by Twilio
        when using API authentication.

        :return: The value of the 'secret' field from the configuration.
        """
        try:
            return self.configuration.secret
        except AttributeError:
            return None

    def collect(self):
        """Collects all events from the Twilio Monitor Event API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the 7-days of data will be collected.
        """
        # Construct the client based on whether "API key" or "auth token" authentication
        # is configured.
        if self.secret:
            client = Client(self.key, self.secret, self.identity)
        else:
            client = Client(self.identity, self.key)

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.now(timezone.utc) - timedelta(days=7)).strftime(
                "%Y-%m-%dT%H:%M:%S%z"
            )

        # Stream the data from the API, paging as required.
        entries = []
        batch_size = 1000

        try:
            events = client.monitor.events.stream(start_date=self.pointer)
        except TwilioException as err:
            raise RequestFailedException(err)

        for event in events:
            # This is less than ideal, but as EventInstance returns datetime objects
            # for the event_date, and does not expose a way to return the entire set
            # of properties we'll need to break encapsulation to get all of the data
            # mapped into the object.
            candidate = event._properties
            candidate["event_date"] = event.event_date.strftime("%Y-%m-%dT%H:%M:%S%z")

            # Track the event.
            entries.append(candidate)

            # Save batches, and clear the list for the next page.
            if len(entries) == batch_size:
                self.save(entries)
                entries = []

        # Save any records not previously saved, before completion.
        self.save(entries)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the entry filtering processor."""

import json
import os
import unittest
from unittest.mock import patch

from grove.models import ProcessorConfig
from grove.processors import filter_entries
from tests import mocks


class ProcessorPathFilterTestCase(unittest.TestCase):
    """Implements unit tests for the entry filtering processor."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Setup the processor and associated configuration for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

        # Create a mapping compatible with the Github Git log fixtures.
        self.processor = filter_entries.Handler(
            ProcessorConfig(
                name="Filter entries with no actor",
                processor="filter_entries",
                filters=[
                    "actor == null",
                ],
            )
        )

    def test_filter_entries(self):
        """Ensure entry filtering operates as expected."""
        # Load and process the fixture.
        entries = json.load(
            open(os.path.join(self.dir, "fixtures/github/git/001.json"), "r")
        )

        # Firstly, ensure four records exist to begin with.
        self.assertEqual(len(entries), 4)

        records = []
        for entry in entries:
            records.extend(self.processor.process(entry))

        # Ensure records with no actor were dropped.
        self.assertEqual(len(records), 1)
        self.assertIn("actor", records[0])



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Twilio messages connector for Grove."""

from datetime import datetime, timedelta, timezone
from functools import lru_cache
from typing import Any, Dict

from twilio.base.exceptions import TwilioException
from twilio.rest import Client

from grove.connectors import BaseConnector
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException, RequestFailedException


class Connector(BaseConnector):
    NAME = "twilio_messages"
    POINTER_PATH = "date_sent"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    @property
    def secret(self):
        """Fetches the API secret from the configuration.

        This is required as this is a third authentication element required by Twilio
        when using API authentication.

        :return: The value of the 'secret' field from the configuration.
        """
        try:
            return self.configuration.secret
        except AttributeError:
            return None

    @lru_cache(maxsize=512)  # noqa: B019
    def _carrier_lookup(self, number: str):
        """Performs a Carrier lookup via the Twilio API.

        This is wrapped with an lru_cache to try and reduce lookup costs for retries of
        the same number between collections - assuming the runtime supports long lived
        LRU caches!

        :return: A dictionary of carrier information from Twilio.
        """
        carrier = self.client.lookups.v1.phone_numbers(number).fetch(type=["carrier"])

        # Extract only carrier and country fields, as we don't want to track PII here.
        return {
            "country_code": carrier.country_code,
            "carrier": carrier.carrier,
        }

    def collect(self):  # noqa: C901
        """Collects logs of all messages from the Twilio Messages API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the 24-hours of data will be collected.
        """
        # Construct the client based on whether "API key" or "auth token" authentication
        # is configured. The client is defined as an instance variable so convenience
        # methods can access the same client.
        if self.secret:
            self.client = Client(self.key, self.secret, self.identity)
        else:
            self.client = Client(self.identity, self.key)

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to 24-hours ago.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.now(timezone.utc) - timedelta(days=1)).strftime(
                "%a, %d %b %Y %H:%M:%S %z"
            )

        # Our LRU cache only has 512 slots, so during a back fill this may be exhausted.
        # To try and alleviate we'll track results in memory during a single execution,
        # too.
        carrier_information: Dict[str, Any] = {}

        # Stream the data from the API, paging as required.
        entries = []
        batch_size = 1000

        try:
            events = self.client.messages.list(date_sent=self.pointer)
        except TwilioException as err:
            raise RequestFailedException(err)

        for event in events:
            # Redaction and lookup of numbers will be performed based on the direction
            # of the message.
            incoming = False

            if event.status == "received":
                number = event.from_
                incoming = True
            else:
                number = event.to

            # Check results cache for carrier information before performing a lookup,
            # as these are expensive in terms of both time and cost (currently $0.005
            # per lookup).
            if number in carrier_information:
                carrier = carrier_information[number]
            else:
                carrier = self._carrier_lookup(number)
                carrier_information[number] = carrier

            # Fields need to be removed from the logs to protect message contents and
            # PII, so we'll construct the log messages ourselves.
            #
            # The use of RFC2822 format dates is horrible, but this is the format that
            # this API expects :(
            message = {
                "account_sid": event.account_sid,
                "date_sent": event.date_sent.strftime("%a, %d %b %Y %H:%M:%S %z"),
                "status": event.status,
                "sid": event.sid,
            }

            # Sanitise the message for PII based on direction.
            if incoming:
                message["to"] = event.to
                try:
                    message["from"] = f"{event.from_[0:4]}..."
                except IndexError:
                    message["from"] = "..."
            else:
                message["from"] = event.from_
                try:
                    message["to"] = f"{event.to[0:4]}..."
                except IndexError:
                    message["to"] = "..."

            # Track the message.
            entries.append({**message, **carrier})

            # Save batches, and clear the list for the next page.
            if len(entries) == batch_size:
                self.save(entries)
                entries = []

        # Save any records not previously saved, before completion.
        self.save(entries)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Okta SystemLog connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.okta.api import Client
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "okta_system_log"
    POINTER_PATH = "published"
    LOG_ORDER = CHRONOLOGICAL

    @property
    def domain(self):
        """Fetches the Okta domain from the configuration.

        This field is used to allow configuration of collection of log data from
        specific Okta domains, including okta-emea.com, and oktapreview.com. This must
        not include the customer name / organisation name.  The default is 'okta.com'.

        :return: The "domain" portion of the connector's configuration.
        """
        try:
            return self.configuration.domain
        except AttributeError:
            return "okta.com"

    def collect(self):
        """Collects all logs from the Okta Audit API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(identity=self.identity, token=self.key, domain=self.domain)
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago.
        now = datetime.utcnow()
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (now - timedelta(days=7)).isoformat(
                sep="T",
                timespec="milliseconds",
            ) + "Z"

        # Get log data from the upstream API, paging if required.
        while True:
            log = client.get_audit_logs(since=self.pointer, cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove cache handlers."""

import abc
from typing import Optional


class BaseCache(abc.ABC):
    @abc.abstractmethod
    def get(self, pk: str, sk: str) -> str:
        """Gets the value for the given key from the cache.

        If the implementation does not differentiate partition and sort keys, these
        values should be combined in an appropriate way to form a cache key.

        :param pk: Partition key of the value to retrieve.
        :param sk: Sort key of the value to retrieve.

        :return: The value from the cache.
        """
        pass

    @abc.abstractmethod
    def set(
        self,
        pk: str,
        sk: str,
        value: str,
        not_set: bool = False,
        constraint: Optional[str] = None,
    ):
        """Stores the value for the given key in a cache.

        If the implementation does not differentiate partition and sort keys, these
        values should be combined in an appropriate way to form a cache key.

        :param pk: Partition key of the value to store.
        :param sk: Sort key of the value to store.
        :param value: Value to store.
        :param not_set: Specifies whether the value must not already be set in the cache
            for the set to be successful.
        :param constraint: An optional condition to use set operation. If provided,
            the currently cached value must match for the delete to be successful.

        :raises ValueError: An incompatible set of parameters were provided.
        :raises DataFormatException: The provided constraint was not satisfied.
        """
        pass

    @abc.abstractmethod
    def delete(self, pk: str, sk: str, constraint: Optional[str] = None):
        """Deletes an entry with the given key from the cache.

        If the implementation does not differentiate partition and sort keys, these
        values should be combined in an appropriate way to form a cache key.

        :param pk: Partition key of the value to delete.
        :param sk: Sort key of the value to delete.
        :param constraint: An optional condition to use during the delete. The value
            provided as the condition must match for the delete to be successful.

        :raises DataFormatException: The provided constraint was not satisfied.
        """
        pass



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the configuration helpers."""

import json
import unittest

import boto3
from moto import mock_ssm

from grove.configs import aws_ssm
from grove.models import ConnectorConfig


class ConfigsAWSSSMTestCase(unittest.TestCase):
    """Implements tests for the Base connector."""

    @mock_ssm
    def test_get_valid_configuration(self):
        """Ensures that the SSM get method is working as expected."""
        fixtures = {
            "/grove/connectors/example/something": {
                "name": "test_one",
                "key": "SUPER_SECRET_ONE",
                "identity": "AAAAAAAA",
                "connector": "first_connector",
            },
            "/grove/connectors/example/some_other_thing": {
                "name": "test_two",
                "key": "SUPER_SECRET_TWO",
                "identity": "BBBBBBBB",
                "connector": "second_connector",
            },
        }

        expected = [
            ConnectorConfig(
                name="test_one",
                key="SUPER_SECRET_ONE",
                identity="AAAAAAAA",
                connector="first_connector",
            ),
            ConnectorConfig(
                name="test_two",
                key="SUPER_SECRET_TWO",
                identity="BBBBBBBB",
                connector="second_connector",
            ),
        ]

        # Push fixtures into the mock SSM.
        client = boto3.client("ssm", region_name="us-east-1")

        for name, fixture in fixtures.items():
            client.put_parameter(
                Name=name, Value=json.dumps(fixture), Type="SecureString"
            )

        # Fetch connector configuration documents via our handler, and ensure they are
        # as we expect.
        configurations = aws_ssm.Handler().get("/")

        for index, configuration in enumerate(configurations):
            self.assertEqual(expected[index], configuration)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Torq API client."""

import base64
import logging
from typing import Any, Dict, Optional

import requests

from grove.exceptions import RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://api.torq.io/public"
API_PAGE_SIZE = 100


class Client:
    def __init__(
        self,
        base_url: Optional[str] = API_BASE_URI,
        identity: Optional[str] = None,
        key: Optional[str] = None,
        bearer_token: Optional[str] = None,
        retry: Optional[bool] = True,
    ) -> None:
        """Setup a new client.

        :param base_url: The Torq api url - default is https://api.tor.io/public
        :param identity: The Torq integration client id
        :param key: The Torq integration client secret
        :param bearer_token: The allocated bearer token provided by Torq given our
            client id and client secret
        :param retry: Whether to automatically retry if recoverable errors are
            encountered, such as rate-limiting.
        """
        self.base_url = base_url
        self.key = key
        self.identity = identity
        self.bearer_token = bearer_token
        self.retry = retry
        self.logger = logging.getLogger(__name__)
        self._set_default_headers()

        if self.bearer_token is None:
            self.refresh_bearer_token()

    def _set_default_headers(self):
        self.headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

    def _get(
        self, url: str, params: Optional[Dict[str, Optional[str]]] = None
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: A URL to perform the HTTP GET against.
        :param parameters: An optional set of HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        try:
            response = requests.get(url, headers=self.headers, params=params)
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def _post(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Optional[str]]] = None,
    ) -> Dict[str, Any]:
        """A POST wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP POST against.
        :param headers: Dictionary of headers to add to the request.
        :param data: HTTP parameters to add to the request.

        :raises RequestFailedException: An HTTP request failed.

        :return: The response to the request in JSON.
        """
        try:
            response = requests.post(
                url,
                headers=headers,
                data=data,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise RequestFailedException(err)

        return response.json()

    def refresh_bearer_token(self):
        """Use Basic Auth to get Bearer Token

        :returns: If the request is successful, the bearer token is returned to
            the Client class header.
        """
        auth_url = "https://auth.torq.io/v1/auth/token"

        # set basic auth value for authorisation header
        basic_auth = str(
            base64.b64encode(bytes(f"{self.identity}:{self.key}", "utf-8")),
            "utf-8",
        )

        bearer_response = self._post(
            auth_url,
            data={
                "grant_type": "client_credentials",
            },
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
                "Authorization": f"Basic {basic_auth}",
            },
        )

        self.bearer_token = bearer_response.get("access_token")

        if self.headers is None or self.headers == {}:
            self._set_default_headers()

        self.headers["Authorization"] = f"Bearer {self.bearer_token}"

    def get_logs(
        self,
        endpoint: str,
        result_field: str,
        start_time: Optional[str] = None,
        limit: Optional[int] = API_PAGE_SIZE,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of logs from Torq which match the provided filters.

        :param start_time: The required date and time of the earliest log entry. Start
            times are in RFC3339 format, for example, 2022-03-09T08:40:18.490771179Z.
        :param result_field: The key name for the list of logs in the returned json.
        :param to_date: The required date and time in UTC of the latest log entry.
        :param limit: The maximum number of items to include in a single response.
        :param cursor: The cursor to use when paging.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        result = self._get(
            f"{self.base_url}/{endpoint}",
            params={
                "start_time": start_time,
                "page_size": str(limit),
                "page_token": cursor,
            },
        )

        # get the list of logs from the result_field parameters value
        data = result.body.get(result_field, [])

        # keep paging until we meet the total number of results
        cursor = result.body.get("next_page_token", None)
        if cursor == "":
            cursor = None

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=data)

    def get_audit_logs(
        self, start_time: Optional[str] = None, cursor: Optional[str] = None
    ) -> AuditLogEntries:
        """Fetches a list of audit logs from Torq which match the provided filters.

        :param start_time: The required date and time of the earliest log entry. Start
            times are in RFC 3339 format, for example, 2022-03-09T08:40:18.490771179Z.
        :param to_date: The required date and time in UTC of the latest log entry.
        :param cursor: The cursor to use when paging.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_logs(
            endpoint="v1alpha/audit_logs",
            result_field="audit_logs",
            start_time=start_time,
            cursor=cursor,
        )

    def get_activity_logs(
        self, start_time: Optional[str] = None, cursor: Optional[str] = None
    ) -> AuditLogEntries:
        """Fetches a list of activity logs from Torq which match the provided filters.

        :param start_time: The required date and time of the earliest log entry. Start
            times are in RFC 3339 format, for example, 2022-03-09T08:40:18.490771179Z.
        :param to_date: The required date and time in UTC of the latest log entry.
        :param cursor: The cursor to use when paging.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_logs(
            endpoint="v1alpha/activity_logs",
            result_field="activity_logs",
            start_time=start_time,
            cursor=cursor,
        )
